/*----------------------------------------------------------------------
  File    : tfscan.c
  Contents: table file scanner management
  Author  : Christian Borgelt
  History : 04.01.1998 file created
            11.03.1998 additional character flags enabled
            12.08.1998 function tfs_copy added
            01.09.1998 several assertions added
            27.09.1998 function tfs_getfld improved
            21.10.1998 bug in tfs_sgetc removed
            26.11.1998 some function parameters changed to const
            04.02.1999 long int changed to int
            16.11.1999 number of characters cleared for an empty field
            01.12.2000 '\r' made a default blank character
            14.07.2001 tfs_sgetc modified, tfs_buf and tfs_err added
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "tfscan.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- functions --- */
#define isblank(c)   tfs_istype(tfs, TFS_BLANK,  c)
#define isfldsep(c)  tfs_istype(tfs, TFS_FLDSEP, c)
#define isrecsep(c)  tfs_istype(tfs, TFS_RECSEP, c)
#define issep(c)     tfs_istype(tfs, TFS_FLDSEP|TFS_RECSEP, c)

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

TFSCAN* tfs_create (void)
{                               /* --- create a table file scanner */
  TFSCAN *tfs;                  /* created table file scanner */
  int    i;                     /* loop variable */
  char   *p;                    /* to traverse character flags */

  tfs = (TFSCAN*)malloc(sizeof(TFSCAN));
  if (!tfs) return NULL;        /* allocate memory */
  for (p = tfs->cflags +256, i = 256; --i >= 0; )
    *--p = '\0';                /* clear and initialize */
  tfs->cflags['\n'] = TFS_RECSEP;    /* character flags */
  tfs->cflags['\t'] = tfs->cflags[' '] = TFS_BLANK|TFS_FLDSEP;
  tfs->cflags['\r'] = TFS_BLANK;
  return tfs;                   /* return created table file scanner */
}  /* tfs_create() */

/*--------------------------------------------------------------------*/

TFSCAN* tfs_dup (const TFSCAN *tfs)
{                               /* --- duplicate a table file scanner */
  TFSCAN *dup;                  /* created duplicate */

  dup = (TFSCAN*)malloc(sizeof(TFSCAN));
  if (!dup) return NULL;        /* create a new table file scanner */
  tfs_copy(dup, tfs);           /* and copy source into it */
  return dup;                   /* return created duplicate */
}  /* tfs_dup() */

/*--------------------------------------------------------------------*/

void tfs_copy (TFSCAN *dst, const TFSCAN *src)
{                               /* --- copy a table file scanner */
  int  i;                       /* loop variable */
  char *d; const char *s;       /* to traverse the character flags */

  assert(src && dst);           /* check arguments */
  s = src->cflags +256; d = dst->cflags +256;
  for (i = 256; --i >= 0; ) *--d = *--s;
}  /* tfs_copy() */             /* copy character flags */

/*--------------------------------------------------------------------*/

int tfs_sgetc (TFSCAN *tfs, const char *s)
{                               /* --- get character from string */
  int c, code;                  /* character and character code */

  if (s) tfs->s = s;            /* if a new string is given, note it */
  if (*tfs->s == '\0')          /* if at the end of the old string, */
    return -1;                  /* abort the function */
  c = (unsigned char)*tfs->s++; /* get the next character */
  if (c != '\\')                /* if no quoted character, */
    return c;                   /* simply return the character */
  c = (unsigned char)*tfs->s++; /* get the next character */
  switch (c) {                  /* and evaluate it */
    case 'a': return '\a';      /* 0x07 (BEL) */
    case 'b': return '\b';      /* 0x08 (BS)  */
    case 'f': return '\f';      /* 0x0c (FF)  */
    case 'n': return '\n';      /* 0x0a (NL)  */
    case 'r': return '\r';      /* 0x0d (CR)  */
    case 't': return '\t';      /* 0x09 (HT)  */
    case 'v': return '\v';      /* 0x0b (VT)  */
    case '0': case '1': case '2': case '3':
    case '4': case '5': case '6': case '7':
      code = c -'0';            /* --- octal character code */
      c    = *tfs->s;           /* get the next character */
      if ((c >= '0') && (c <= '7')) code = (code << 3) +c -'0';
      else return code;         /* decode second digit */
      c    = *++tfs->s;         /* get the next character */
      if ((c >= '0') && (c <= '7')) code = (code << 3) +c -'0';
      else return c;            /* decode third digit */
      tfs->s++;                 /* consume the decoded character */
      return code & 0xff;       /* and return the character code */
    case 'x':                   /* --- hexadecimal character code */
      c = *tfs->s;              /* get the next character */
      if      ((c >= '0') && (c <= '9')) code = c -'0';
      else if ((c >= 'a') && (c <= 'f')) code = c -'a' +10;
      else if ((c >= 'A') && (c <= 'F')) code = c -'A' +10;
      else return 'x';          /* decode first digit */
      c = *++tfs->s;            /* get the next character */
      if      ((c >= '0') && (c <= '9')) code = (code << 4) +c -'0';
      else if ((c >= 'a') && (c <= 'f')) code = (code << 4) +c -'a' +10;
      else if ((c >= 'A') && (c <= 'F')) code = (code << 4) +c -'A' +10;
      else return code;         /* decode second digit */
      tfs->s++;                 /* consume the decoded character */
      return code;              /* and return the character code */
    default:                    /* non-function characters */
      if (*tfs->s == '\0') return '\\';
      else                 return (unsigned char)*tfs->s++;
  }                             /* return character or backslash */
}  /* tfs_sgetc() */

/*--------------------------------------------------------------------*/

int tfs_chars (TFSCAN *tfs, int type, const char *chars)
{                               /* --- set characters */
  int  i, c, d;                 /* loop variable, characters */
  char *p;                      /* to traverse character flags */

  assert(tfs);                  /* check argument */
  if (!chars) return -1;        /* if no characters given, abort */
  p = tfs->cflags +256;         /* clear character flags in type */
  for (i = 256; --i >= 0; ) *--p &= (char)~type;
  for (c = d = tfs_sgetc(tfs, chars); c >= 0; c = tfs_sgetc(tfs, NULL))
    tfs->cflags[c] |= (char)type;  /* set character flags */
  return (d >= 0) ? d : 0;      /* return first character */
}  /* tfs_chars() */

/*--------------------------------------------------------------------*/

int tfs_getfld (TFSCAN *tfs, FILE *file, char *buf, int len)
{                               /* --- read a table field */
  int  c;                       /* character read */
  int  d;                       /* delimiter type */
  char *p;                      /* to traverse the buffer */

  assert(tfs && file && (!buf || (len >= 0)));
  if (!buf) {                   /* if no buffer given, use internal */
    buf = tfs->buf; len = TFS_SIZE; }
  p = buf; *p = '\0';           /* clear read buffer and */
  tfs->cnt = 0;                 /* the number of characters read */
  do {                          /* --- skip leading blanks */
    c = getc(file);             /* get next character */
    if (c == EOF) return (ferror(file)) ? -1 : TFS_EOF;
  } while (isblank(c));         /* while character is blank */
  if (issep(c))                 /* check for field/record separator */
    return (isfldsep(c)) ? TFS_FLD : TFS_REC;

  if (c == '"') {               /* if the field is a "" delimited field */
    do {
      c = getc(file);
      if (c == EOF) return (ferror(file)) ? -1 : TFS_EOF;
    } while (isblank(c));
    while (1) {
      if (c == '"') break;      /* end of field reached */
      if (isrecsep(c)) { d = TFS_REC; break; }
      if (c == EOF) { d = (ferror(file)) ? -1 : TFS_EOF; break; }
      if (len >= 0) {           /* if buffer is not full, */
	len--; *p++ = (char)c; }/* store character in buffer */
      c = getc(file);
    }
    if (c == '"') {             
      while (1) {               /* read to the next field seperator */
	c = getc(file);
	if (!isblank(c)) { d = (c == EOF) ? TFS_EOF : -1; break; }
	if (issep(c)) { d = (isfldsep(c))  ? TFS_FLD : TFS_REC; break; }
	if (c == EOF) { d = (ferror(file)) ? -1 : TFS_EOF; break; }
      }
    }
  } 
  else {                        /* field does not start with a '"' */
    while (1) {                   /* --- read value */
      if (len >= 0) {             /* if buffer is not full, */
	len--; *p++ = (char)c; }  /* store character in buffer */
      c = getc(file);             /* get next character */
      if (issep(c)) { d = (isfldsep(c))  ? TFS_FLD : TFS_REC; break; }
      if (c == EOF) { d = (ferror(file)) ? -1      : TFS_EOF; break; }
    }                             /* while character is no separator */
  }
  while (isblank(*--p));        /* --- remove trailing blanks */
  *++p = '\0';                  /* terminate string in buffer */
  tfs->cnt = (int)(p -buf);     /* store number of characters read */
  if (d != TFS_FLD) return d;   /* if not at field separator, abort */
  while (isblank(c)) {          /* --- skip trailing blanks */
    c = getc(file);             /* get next character */
    if (c == EOF) return (ferror(file)) ? -1 : TFS_EOF;
  }                             /* check for end of file */
  if (isrecsep(c))  return TFS_REC;  /* check for record separator */
  if (!isfldsep(c)) ungetc(c, file); /* put back character */
  return TFS_FLD;               /* return delimiter type */
}  /* tfs_getfld() */
